"""contains packages related to SCADA protocol layers."""

from scapy.contrib.scada.iec104 import *  # noqa F403,F401

# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Thomas Tannhaeuser <hecke@naberius.de>
# This program is published under a GPLv2 license
#
# scapy.contrib.description = SCADA related layers
# scapy.contrib.status = loads


# Package of contrib SCADA modules.
