# This file is part of Scapy
# Copyright (C) 2007, 2008, 2009 Arnaud Ebalard
#                     2015, 2016 Maxence Tury
# This program is published under a GPLv2 license

"""
Cryptographic capabilities for TLS.
"""
