# This file is part of Scapy
# Copyright (C) 2007, 2008, 2009 Arnaud Ebalard
#               2015, 2016, 2017 Maxence Tury
# This program is published under a GPLv2 license

"""
Aggregate some TLS crypto objects.
"""

# flake8: noqa: F403

from scapy.layers.tls.crypto.suites import *  # noqa: F401
